commitID=$(git rev-parse --short HEAD)
echo "Moving tag"
hub push origin :refs/tags/nightly
git tag -f nightly $commitID
hub push --tags

MESSAGE=$'Nightly build\n\nAutogenerated nightly build of micro'

echo "Creating new release"
hub release create nightly \
    --prerelease \
    --draft=false \
    --message "$MESSAGE."
